# pylint: disable=line-too-long
"""
This module provides a reporter with JUnit XML output.

Mapping of behave_bo model elements to XML elements::

    feature     -> xml_element:testsuite
    scenario    -> xml_element:testcase

XML document structure::

    # -- XML elements:
    # CARDINALITY SUFFIX:
    #   ?   optional (zero or one)
    #   *   many0 (zero or more)
    #   +   many (one or more)
    testsuites := sequence<testsuite>
    testsuite:
        properties? : sequence<property>
        testcase* :
            error?      : text
            failure?    : text
            system-out  : text
            system-err  : text

    testsuite:
        @name       : TokenString
        @tests      : int
        @failures   : int
        @errors     : int
        @skipped    : int
        @time       : Decimal       # Duration in seconds
        # -- SINCE: behave_bo-1.2.6
        @timestamp  : IsoDateTime
        @hostname   : string

    testcase:
        @name       : TokenString
        @classname  : TokenString
        @status     : string        # Status enum
        @time       : Decimal       # Elapsed seconds

    error:
        @message    : string
        @type       : string

    failure:
        @message    : string
        @type       : string

    # -- HINT: Not used
    property:
        @name  : TokenString
        @value : string

    type Status : Enum("passed", "failed", "skipped", "untested")

Note that a spec for JUnit XML output was not clearly defined.
Best sources are:

* `JUnit XML`_ (for PDF)
* JUnit XML (`ant spec 1`_, `ant spec 2`_)


.. _`JUnit XML`:  http://junitpdfreport.sourceforge.net/managedcontent/PdfTranslation
.. _`ant spec 1`: https://github.com/windyroad/JUnit-Schema
.. _`ant spec 2`: http://svn.apache.org/repos/asf/ant/core/trunk/src/main/org/apache/tools/ant/taskdefs/optional/junit/XMLJUnitResultFormatter.java
"""
# pylint: enable=line-too-long

import codecs
import os
import os.path
import re
import traceback
from datetime import (
    datetime,
)
from functools import (
    partial,
)
from xml.etree import (
    ElementTree,
)

import requests
from django.template import (
    Template,
    Context,
)
from lxml import (
    etree,
)

from behave_bo.consts import (
    scenario_group_prefix,
)
from behave_bo.formatter import (
    ansi_escapes,
)
from behave_bo.model import (
    Scenario,
    ScenarioOutline,
    Step,
)
from behave_bo.model_core import (
    Status,
)
from behave_bo.model_describe import (
    ModelDescriptor,
)
from behave_bo.reporter.base import (
    Reporter,
)
from behave_bo.textutil import (
    indent,
    make_indentation,
    text as _text,
)
from behave_bo.userdata import (
    UserDataNamespace,
)


def CDATA(text=None):   # pylint: disable=invalid-name
    # -- issue #70: remove_ansi_escapes(text)
    element = ElementTree.Element('![CDATA[')
    element.text = ansi_escapes.strip_escapes(text)
    return element


class ElementTreeWithCDATA(ElementTree.ElementTree):
    # pylint: disable=redefined-builtin, no-member
    def _write(self, file, node, encoding, namespaces):
        """This method is for ElementTree <= 1.2.6"""

        if node.tag == '![CDATA[':
            text = node.text.encode(encoding)
            file.write(f"\n<![CDATA[{text}]]>\n")
        else:
            ElementTree.ElementTree._write(self, file, node, encoding,
                                           namespaces)


if hasattr(ElementTree, '_serialize'):
    # pylint: disable=protected-access

    def _serialize_xml3(write, elem, qnames, namespaces,
                        short_empty_elements=None,
                        orig=ElementTree._serialize_xml):
        if elem.tag == '![CDATA[':
            write("\n<{tag}{text}]]>\n".format(
                tag=elem.tag, text=elem.text))
            return
        if short_empty_elements:
            # python >=3.3
            return orig(write, elem, qnames, namespaces, short_empty_elements)
        else:
            # python <3.3
            return orig(write, elem, qnames, namespaces)

    ElementTree._serialize_xml = ElementTree._serialize['xml'] = _serialize_xml3


class FeatureReportData:
    """
    Provides value object to collect JUnit report data from a Feature.
    """
    def __init__(self, feature, filename, classname=None):
        if not classname and filename:
            classname = filename.replace('/', '.')
        self.feature = feature
        self.filename = filename
        self.classname = classname
        self.testcases = []
        self.counts_tests = 0
        self.counts_errors = 0
        self.counts_failed = 0
        self.counts_skipped = 0

    def reset(self):
        self.testcases = []
        self.counts_tests = 0
        self.counts_errors = 0
        self.counts_failed = 0
        self.counts_skipped = 0


class JUnitReporter(Reporter):
    """Generates JUnit-like XML test report for behave_bo.
    """
    # -- XML REPORT:
    userdata_scope = "behave_bo.reporter.junit"
    show_timings = True     # -- Show step timings.
    show_skipped_always = False
    show_timestamp = True
    show_hostname = True
    # -- XML REPORT PART: Describe scenarios
    show_scenarios = ''   # Show particular scenarios descriptions.
    show_all_scenarios = True   # Show all scenarios descriptions.
    show_tags = True
    show_multiline = True
    build_url = ''

    def __init__(self, config):
        super().__init__(config)
        self.rerun = False
        self._features_failed_suites = {}
        self.setup_with_userdata(config.userdata)

    def setup_with_userdata(self, userdata):
        """Setup JUnit reporter with userdata information.
        A user can now tweak the output format of this reporter.

        EXAMPLE:
        .. code-block:: ini

            # -- FILE: behave_bo.ini
            [behave_bo.userdata]
            behave_bo.reporter.junit.show_hostname = false
        """
        # -- EXPERIMENTAL:
        config = UserDataNamespace(self.userdata_scope, userdata)
        self.show_hostname = config.getbool("show_hostname", self.show_hostname)
        self.show_multiline = config.getbool("show_multiline", self.show_multiline)
        self.show_all_scenarios = config.getbool("show_all_scenarios", self.show_all_scenarios)

        show_scenarios_str = config.get("show_scenarios", '')
        if show_scenarios_str:
            self.show_scenarios = show_scenarios_str.split(',')
        else:
            self.show_scenarios = []

        self.show_tags = config.getbool("show_tags", self.show_tags)
        self.show_timings = config.getbool("show_timings", self.show_timings)
        self.show_timestamp = config.getbool("show_timestamp", self.show_timestamp)
        self.show_skipped_always = config.getbool("show_skipped_always", self.show_skipped_always)

        if 'build_url' in userdata:
            self.build_url = userdata['build_url']

    def make_feature_filename(self, feature):
        """Пропатчено для формирования короткого имени файла в отчете.

        А в случае параллельного запуска тестов, если в тегах передан тег группы - добавляет к имени фичи постфикс
        чтобы потом собрать из нескольких файлов JUNIT-xml для feature - один файл JUNIT-xml.

        Args:
            self: Объект класса JUnitReporter.
            feature: Объект класса Feature.

        """
        feature_name = feature.location.filename
        base_name = os.path.basename(feature_name).split('.', 1)[0]

        if feature_name.startswith('plugins/'):
            plugin = feature_name.split('/')[1]
            base_name = f'{plugin}.{base_name}'

        if hasattr(self.config, 'group_number'):
            base_name += f'_{scenario_group_prefix}{self.config.group_number}'

        return base_name

    @property
    def show_skipped(self):
        return self.config.show_skipped or self.show_skipped_always

    # -- REPORTER-API:
    def feature(self, feature):
        if feature.status == Status.skipped and not self.show_skipped:
            # -- SKIP-OUTPUT: If skipped features should not be shown.
            return

        feature_filename = self.make_feature_filename(feature)
        classname = feature_filename
        report = FeatureReportData(feature, feature_filename)
        now = datetime.now()
        testcase_by_name = {}

        suite = ElementTree.Element('testsuite')
        feature_name = (feature.name or feature_filename).replace('.', '')
        suite.set('name', f'{classname}.{feature_name}')

        # -- BUILD-TESTCASES: From scenarios
        for scenario in feature:
            if isinstance(scenario, ScenarioOutline):
                scenario_outline = scenario
                self._process_scenario_outline(scenario_outline, report)
            else:
                self._process_scenario(scenario, report)

        # -- ADD TESTCASES to testsuite:
        for testcase in report.testcases:
            suite.append(testcase)

            if not self.config.parallel and self.rerun:
                testcase_by_name[testcase.attrib[JunitXMLAttrsEnum.NAME]] = testcase

        suite.set('tests', _text(report.counts_tests))
        suite.set('errors', _text(report.counts_errors))
        suite.set('failures', _text(report.counts_failed))
        suite.set('skipped', _text(report.counts_skipped))  # WAS: skips
        suite.set('time', _text(round(feature.duration, 6)))
        # -- SINCE: behave_bo-1.2.6.dev0
        if self.show_timestamp:
            suite.set('timestamp', _text(now.isoformat()))
        if self.show_hostname:
            suite.set('hostname', _text(gethostname()))

        if self.build_url:
            suite.set('build_url', self.build_url)

        if not os.path.exists(self.config.junit_directory):
            # -- ENSURE: Create multiple directory levels at once.
            os.makedirs(self.config.junit_directory)

        if not self.config.parallel:
            if feature_name in self._features_failed_suites and self.rerun:
                prev_suite = self._features_failed_suites[feature_name]
                prev_suite.set('failures', _text(suite.attrib['failures']))

                for tc in prev_suite:
                    if tc.attrib[JunitXMLAttrsEnum.STATUS] != JunitXMLAttrsEnum.FAILED:
                        continue

                    rerun_tc = testcase_by_name[tc.attrib[JunitXMLAttrsEnum.NAME]]

                    for attr in (
                        JunitXMLAttrsEnum.STATUS,
                        JunitXMLAttrsEnum.TIME,
                        JunitXMLAttrsEnum.TIME_PROCESS,
                        JunitXMLAttrsEnum.PID,
                    ):
                        tc.attrib[attr] = rerun_tc.attrib[attr]

                suite = prev_suite

            elif report.counts_failed > 0:
                self._features_failed_suites[feature_name] = suite

        tree = ElementTreeWithCDATA(suite)
        report_dirname = self.config.junit_directory
        report_basename = f'TESTS-{feature_filename}.xml'
        report_filename = os.path.join(report_dirname, report_basename)
        tree.write(codecs.open(report_filename, "wb"), "UTF-8")

    # -- MORE:
    # pylint: disable=line-too-long
    @staticmethod
    def select_step_with_status(status, steps):
        """Helper function to find the first step that has the given
        step.status.

        EXAMPLE: Search for a failing step in a scenario (all steps).
            >>> scenario = ...
            >>> failed_step = select_step_with_status(Status.failed, scenario)
            >>> failed_step = select_step_with_status(Status.failed, scenario.all_steps)
            >>> assert failed_step.status == Status.failed

        EXAMPLE: Search only scenario steps, skip background steps.
            >>> failed_step = select_step_with_status(Status.failed, scenario.steps)

        :param status:  Step status to search for (as enum value).
        :param steps:   List of steps to search in (or scenario).
        :returns: Step object, if found.
        :returns: None, otherwise.

        .. versionchanged:: 1.2.6
            status: Use enum value instead of string (or string).
        """
        for step in steps:
            assert isinstance(step, Step), f"TYPE-MISMATCH: step.class={step.__class__.__name__}"
            if step.status == status:
                return step
        # -- OTHERWISE: No step with the given status found.
        # KeyError("Step with status={0} not found".format(status))
        return None
    # pylint: enable=line-too-long

    def describe_step(self, step):
        status_text = _text(step.status.name)
        if self.show_timings:
            status_text += f' in {step.duration:.3f}s (pt {step.duration_pt:.3f}s)'
        text = f'{step.keyword} {step.name} ... '
        text += f'{status_text}\n'
        if self.show_multiline:
            prefix = make_indentation(2)
            if step.text:
                text += ModelDescriptor.describe_docstring(step.text, prefix)
            elif step.table:
                text += ModelDescriptor.describe_table(step.table, prefix)
        return text

    @classmethod
    def describe_tags(cls, tags):
        text = ''
        if tags:
            text = '@' + ' @'.join(tags)
        return text

    def describe_scenario(self, scenario):
        """Describe the scenario and the test status.
        NOTE: table, multiline text is missing in description.

        :param scenario:  Scenario that was tested.
        :return: Textual description of the scenario.
        """
        header_line = '\n@scenario.begin\n'
        if self.show_tags and scenario.tags:
            header_line += f'\n  {self.describe_tags(scenario.tags)}\n'

        header_line += f'  {scenario.keyword}: {scenario.name}\n'
        footer_line = '\n@scenario.end\n' + '-' * 80 + '\n'
        text = ''
        for step in scenario:
            text += self.describe_step(step)
        step_indentation = make_indentation(4)
        return header_line + indent(text, step_indentation) + footer_line

    def _process_scenario(self, scenario, report):
        """Process a scenario and append information to JUnit report object.
        This corresponds to a JUnit testcase:

          * testcase.@classname = f(filename) +'.'+ feature.name
          * testcase.@name   = scenario.name
          * testcase.@status = scenario.status
          * testcase.@time   = scenario.duration

        Distinguishes now between failures and errors.
        Failures are AssertationErrors: expectation is violated/not met.
        Errors are unexpected RuntimeErrors (all other exceptions).

        If a failure/error occurs, the step, that caused the failure,
        and its location are provided now.

        :param scenario:  Scenario to process.
        :param report:    Context object to store/add info to (outgoing param).
        """
        # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        assert isinstance(scenario, Scenario)
        assert not isinstance(scenario, ScenarioOutline)
        if scenario.status != Status.skipped or self.show_skipped:
            # -- NOTE: Count only if not-skipped or skipped should be shown.
            report.counts_tests += 1
        classname = report.classname
        feature = report.feature
        feature_name = feature.name
        if not feature_name:
            feature_name = self.make_feature_filename(feature)

        feature_name = feature_name.replace('.', '')

        case = ElementTree.Element('testcase')
        case.set("classname", f"{classname}.{feature_name}")
        case.set("name", scenario.name_w_tag or "")
        case.set("status", scenario.status.name)
        case.set("time", _text(round(scenario.duration, 6)))
        case.set("time_process", _text(round(scenario.duration_pt, 6)))
        case.set("pid", f'{os.getpid()}')

        step = None
        failing_step = None
        if scenario.status == Status.failed:
            for status in (Status.failed, Status.undefined):
                step = self.select_step_with_status(status, scenario)
                if step:
                    break
            # -- NOTE: Scenario may fail now due to hook-errors.
            element_name = "failure"
            if step and isinstance(step.exception, (AssertionError, type(None))):
                # -- FAILURE: AssertionError
                assert step.status in (Status.failed, Status.undefined)
                report.counts_failed += 1
            else:
                # -- UNEXPECTED RUNTIME-ERROR:
                report.counts_errors += 1
                element_name = "error"
            # -- COMMON-PART:
            failure = ElementTree.Element(element_name)
            if step:
                step_text = self.describe_step(step).rstrip()
                text = f"\nFailing step: {step_text}\nLocation: {step.location}\n"
                message = _text(step.exception)
                failure.set('type', step.exception.__class__.__name__)
                failure.set('message', message)
                text += _text(step.error_message)
            else:
                # -- MAYBE: Hook failure before any step is executed.
                failure_type = "UnknownError"
                if scenario.exception:
                    failure_type = scenario.exception.__class__.__name__
                failure.set('type', failure_type)
                failure.set('message', scenario.error_message or "")
                traceback_lines = traceback.format_tb(scenario.exc_traceback)
                traceback_lines.insert(0, "Traceback:\n")
                text = _text("".join(traceback_lines))
            failure.append(CDATA(text))
            case.append(failure)
        elif (scenario.status in (Status.skipped, Status.untested)
              and self.show_skipped):
            report.counts_skipped += 1
            step = self.select_step_with_status(Status.undefined, scenario)
            if step:
                # -- UNDEFINED-STEP:
                report.counts_failed += 1
                failure = ElementTree.Element("failure")
                failure.set("type", "undefined")
                failure.set("message", f"Undefined Step: {step.name}")
                case.append(failure)
            else:
                skip = ElementTree.Element('skipped')
                case.append(skip)

        # Create stdout section for each test case
        stdout = ElementTree.Element("system-out")
        text = ""
        if self.show_all_scenarios or scenario.main_tag in self.show_scenarios:
            text = self.describe_scenario(scenario)

        # Append the captured standard output
        if scenario.captured.stdout:
            output = _text(scenario.captured.stdout)
            text += f"\nCaptured stdout:\n{output}\n"
        stdout.append(CDATA(text))
        case.append(stdout)

        # Create stderr section for each test case
        if scenario.captured.stderr:
            stderr = ElementTree.Element("system-err")
            output = _text(scenario.captured.stderr)
            text = f"\nCaptured stderr:\n{output}\n"
            stderr.append(CDATA(text))
            case.append(stderr)

        if scenario.status != Status.skipped or self.show_skipped:
            report.testcases.append(case)

    def _process_scenario_outline(self, scenario_outline, report):
        assert isinstance(scenario_outline, ScenarioOutline)
        for scenario in scenario_outline:
            assert isinstance(scenario, Scenario)
            self._process_scenario(scenario, report)


class JunitXMLAttrsEnum:
    """
    Перечисление атрибутов тега testsuite
    """

    ERRORS_COUNT = 'errors'
    FAILURES_COUNT = 'failures'
    SKIPPED_COUNT = 'skipped'
    TESTS_COUNT = 'tests'
    TIME = 'time'
    TIME_PROCESS = 'time_process'
    PID = 'pid'
    TIMESTAMP = 'timestamp'
    NAME = 'name'
    HOSTNAME = 'hostname'
    BUILD_URL = 'build_url'
    STATUS = 'status'
    UNTESTED = 'untested'
    FAILED = 'failed'

    NUMERIC_ATTRIBUTES = (
        ERRORS_COUNT,
        FAILURES_COUNT,
        SKIPPED_COUNT,
        TESTS_COUNT,
        TIME,
        TIME_PROCESS,
    )


class JunitReportsCombiner:
    """
    Объединитель частей Junit-отчётов в один файл.
    """

    def __init__(
        self,
        reports_dir_path,
        html_report_destination_path=None,
        tests_execution_start_time=None,
        rerun=False,
    ) -> None:
        """
        @param tests_execution_start_time: Время начала выполнения тестов.
            Нужно для расчёта времени выполнения тест-сьюта.
        @param reports_dir_path: путь к директории с xml-файлами junit-отчётов
        """
        super().__init__()

        if tests_execution_start_time:
            self._tests_execution_start_time = tests_execution_start_time
        else:
            self._tests_execution_start_time = datetime.now()

        self._reports_dir_path = reports_dir_path
        self._html_report_destination_path = html_report_destination_path or self._reports_dir_path
        self._rerun = rerun

        self._reports_data = {}
        self._reports_parts_data = {}
        self._reports_files = []
        self._reports_parts_files = []

        self._scenario_group_name_pattern = fr'_{scenario_group_prefix}\d+'
        self._report_part_filename_pattern = fr'(.*?){self._scenario_group_name_pattern}\.xml'  # noqa
        self._report_filename_pattern = fr'(.*?)\.xml'  # noqa

        self._suite_attrs_with_parsers = {
            JunitXMLAttrsEnum.ERRORS_COUNT: int,
            JunitXMLAttrsEnum.FAILURES_COUNT: int,
            JunitXMLAttrsEnum.SKIPPED_COUNT: int,
            JunitXMLAttrsEnum.TESTS_COUNT: int,
            JunitXMLAttrsEnum.HOSTNAME: str,
            JunitXMLAttrsEnum.BUILD_URL: str,
            JunitXMLAttrsEnum.TIMESTAMP: lambda date_str: datetime.fromisoformat(date_str),
            JunitXMLAttrsEnum.NAME: partial(
                re.sub,
                self._scenario_group_name_pattern,
                '',
            ),
        }

        self._test_cases_with_time = []

    def prepare_report_files_paths(self):
        """
        Сформировать набор файлов частей junit-отчётов
        """
        for _, _, files in os.walk(self._reports_dir_path, topdown=False):
            for filename in files:
                match_part = re.match(self._report_part_filename_pattern, filename)
                match_report = re.match(self._report_filename_pattern, filename)

                if match_part:
                    self._reports_parts_files.append(
                        (
                            match_part.group(1),
                            os.path.join(self._reports_dir_path, filename),
                        )
                    )
                elif match_report:
                    self._reports_files.append(
                        (
                            match_report.group(1),
                            os.path.join(self._reports_dir_path, filename),
                        )
                    )

    def collect_reports_data_from_files(self, files, data):
        """
        Получить данные о junit-отчётах из xml-файлов
        """
        for feature_name, path in files:
            tree = etree.parse(path)
            suite_xml_root = tree.getroot()

            if feature_name not in data:
                data[feature_name] = {
                    'suite': {},
                    'test_cases': [],
                }

            test_suite = data[feature_name]['suite']

            for attr, parser in self._suite_attrs_with_parsers.items():
                value = parser(suite_xml_root.attrib.get(attr, 0))

                if attr not in test_suite:
                    test_suite[attr] = value
                elif attr == JunitXMLAttrsEnum.TIMESTAMP:
                    if test_suite[attr] < value:
                        test_suite[attr] = value
                elif attr in JunitXMLAttrsEnum.NUMERIC_ATTRIBUTES:
                    test_suite[attr] += value

            data[feature_name]['test_cases'].extend(
                tree.iter('testcase')
            )

    def prepare_tc_url(self, build_url, test_case):
        """
        Сформировать url-адрес тесткейса в тестовом отчёте.
        """
        feature_file_path = requests.utils.quote(test_case.attrib.get('classname', '').replace('.', '/'))
        scenario_path = re.sub(r'[^\w0-9]', '_', test_case.attrib.get(JunitXMLAttrsEnum.NAME, ''))
        return f'{build_url}testReport/{feature_file_path}/{scenario_path}/'

    def export_reports_parts_data_to_files(self):
        """
        Записать полученные данные о junit-отчётах в xml-файлы junit-отчётов
        """
        reports_paths = []
        name_parser = self._suite_attrs_with_parsers[JunitXMLAttrsEnum.NAME]

        for feature_filename, data in self._reports_parts_data.items():
            suite = etree.Element('testsuite')

            for attr, _ in self._suite_attrs_with_parsers.items():
                suite.set(attr, str(data['suite'].get(attr)))

            last_suite_timestamp = data['suite'].get(JunitXMLAttrsEnum.TIMESTAMP)  # noqa
            duration = last_suite_timestamp - self._tests_execution_start_time
            suite.set(JunitXMLAttrsEnum.TIME, str(duration.total_seconds()))
            build_url = data['suite'].get(JunitXMLAttrsEnum.BUILD_URL)

            for test_case in data['test_cases']:
                if test_case.attrib[JunitXMLAttrsEnum.STATUS] == JunitXMLAttrsEnum.UNTESTED:
                    continue

                raw_classname = test_case.attrib.get('classname', '')

                classname = name_parser(raw_classname)
                test_case.set('classname', classname)

                tc_url = ''

                for child in test_case:
                    child.text = etree.CDATA(child.text)
                    if child.text.strip():
                        tc_url = self.prepare_tc_url(build_url, test_case)

                suite.append(test_case)

                self._test_cases_with_time.append(
                    (
                        round(float(test_case.attrib.get(JunitXMLAttrsEnum.TIME, 0)), 2),
                        round(float(test_case.attrib.get(JunitXMLAttrsEnum.TIME_PROCESS, 0)), 2),
                        test_case.attrib.get(JunitXMLAttrsEnum.NAME, ''),
                        test_case.attrib.get(JunitXMLAttrsEnum.STATUS, ''),
                        test_case.attrib.get(JunitXMLAttrsEnum.PID, ''),
                        tc_url,
                    )
                )

            report_file_path = os.path.join(
                self._reports_dir_path,
                f'{feature_filename}.xml',
            )
            tree = etree.ElementTree(suite)
            tree.write(
                report_file_path,
                encoding='UTF-8',
            )
            reports_paths.append(report_file_path)

        return reports_paths

    def remove_reports_parts_files(self):
        """
        Удаляет файлы частей Junit-отчётов.
        Они не нужны, т.к. были объединены в общий файл junit-отчёта.
        """
        for _, path in self._reports_parts_files:
            if os.path.exists(path):
                os.remove(path)

    def change_reports_parts_rerun_data(self):
        """
        Использует в качестве данных для выгрузки данные из ранее сформированных xml-файлов,
        вносит в них исправления с учётом перезапуска упавших тестов.
        """
        self.collect_reports_data_from_files(
            files=self._reports_files,
            data=self._reports_data,
        )

        for feature_filename, data in self._reports_parts_data.items():
            reports_data = self._reports_data[feature_filename]
            reports_data['suite'][JunitXMLAttrsEnum.FAILURES_COUNT] = data['suite'][JunitXMLAttrsEnum.FAILURES_COUNT]

            for tc in reports_data['test_cases']:
                if tc.attrib[JunitXMLAttrsEnum.STATUS] != JunitXMLAttrsEnum.FAILED:
                    continue

                for rerun_tc in data['test_cases']:
                    if (
                        tc.attrib[JunitXMLAttrsEnum.NAME] == rerun_tc.attrib[JunitXMLAttrsEnum.NAME]
                        and rerun_tc.attrib[JunitXMLAttrsEnum.STATUS] != JunitXMLAttrsEnum.UNTESTED
                    ):
                        for attr in (
                            JunitXMLAttrsEnum.STATUS,
                            JunitXMLAttrsEnum.TIME,
                            JunitXMLAttrsEnum.TIME_PROCESS,
                            JunitXMLAttrsEnum.PID,
                        ):
                            tc.attrib[attr] = rerun_tc.attrib[attr]
                        break

            self._reports_parts_data[feature_filename] = reports_data

    def join_reports(self):
        """
        Собрать данные отчётов из файлов.
        Объединить части junit-отчётов полученные при выполнении параллельных процессов.
        Если это режим перезапуска, внести исправления в junit-отчёты по итогу перезапуска.
        Удалить файлы частей junit-отчётов.
        """
        self.prepare_report_files_paths()
        self.collect_reports_data_from_files(
            files=self._reports_parts_files,
            data=self._reports_parts_data,
        )

        if self._rerun:
            self.change_reports_parts_rerun_data()

        reports_paths = self.export_reports_parts_data_to_files()

        self.remove_reports_parts_files()

        return reports_paths

    def export_top_scenarios_by_time(self):
        if self._rerun:
            # не выполняем формирования файла топ запросов если в режиме перезапуска
            return

        testcases = []

        for time, time_process, tc_name, tc_status, tc_pid, tc_url in sorted(self._test_cases_with_time, key=lambda k: k[0], reverse=True):
            testcases.append({
                'name': tc_name,
                'status': tc_status,
                'pid': tc_pid,
                'url': tc_url,
                'time': time,
                'time_process': time_process,
            })

        with open(os.path.join(os.path.dirname(__file__), 'templates/autotests_report_template.html')) as ft:
            django_template = Template(ft.read())

        rendered_data = django_template.render(Context({
            'title': 'Список автотестов по времени выполнения',
            'page_type': 'top_scenarios',
            'testcases': testcases,
        }))

        with open(os.path.join(self._html_report_destination_path, f'top_scenarios.html'), 'w') as f:
            f.write(rendered_data)

    def export_html_report(self):
        """Формирует html-отчёт из junit-файлов
        """
        self.prepare_report_files_paths()
        self.collect_reports_data_from_files(
            files=self._reports_files,
            data=self._reports_data,
        )

        with open(os.path.join(os.path.dirname(__file__), 'templates/autotests_report_template.html')) as ft:
            django_template = Template(ft.read())

        total_count = 0
        failed_data = []

        for suite_data in self._reports_data.values():
            try:
                suite_name = suite_data['suite']['name'].split('.', 1)[1]
                suite_data['suite']['name'] = suite_name
            except:
                pass

            tested_test_cases = []

            for test_case in suite_data['test_cases']:
                tc_main_tag = test_case.attrib['name'].split(' ', 1)[0]
                test_case.attrib['main_tag'] = tc_main_tag

                if test_case.attrib[JunitXMLAttrsEnum.STATUS] == JunitXMLAttrsEnum.UNTESTED:
                    continue
                elif test_case.attrib[JunitXMLAttrsEnum.STATUS] == JunitXMLAttrsEnum.FAILED:
                    failed_data.append(test_case)
                    try:
                        failure_text = test_case.find('failure').text
                    except:
                        failure_text = 'текст ошибки недоступен'

                    try:
                        failure_text = test_case.find('error').text
                    except:
                        pass

                    test_case.attrib['failure_text'] = failure_text

                total_count += 1
                tested_test_cases.append(test_case)

            suite_data['test_cases'] = tested_test_cases

        now = datetime.now()

        try:
            from web_bb.core.base.helpers import version_info
            version = sorted(version_info(check_installed_apps=False), key=lambda el: el['name'] == 'web_bb_app', reverse=True)
        except Exception as e:
            print(f'Ошибка при формировании информации о версии: {e}')
            version = None

        rendered_data = django_template.render(Context({
            'title': f'Результат автотестирования {now}',
            'page_type': 'autotest_result',
            'total_count': total_count,
            'failed_data': failed_data,
            'suites_data': self._reports_data.values(),
            'version_info': version,
        }))

        export_path = os.path.join(self._html_report_destination_path, f'autotests_report_{now}.html')
        with open(export_path, 'w') as f:
            f.write(rendered_data)

        return export_path

# -----------------------------------------------------------------------------
# SUPPORT:
# -----------------------------------------------------------------------------
def gethostname():
    """Return hostname of local host (as string)"""
    import socket
    return socket.gethostname()



"""Basic types (helper classes)."""

import sys
import traceback


class Unknown:
    """Placeholder for unknown/missing information, distinguishable from None.

    .. code-block:: python

        data = {}
        value = data.get("name", Unknown)
        if value is Unknown:
            # -- DO SOMETHING
            ...
    """


class ExceptionUtil:
    """Provides a utility class for accessing/modifying exception information.

    .. seealso:: PEP-3134 Chained excpetions
    """
    # pylint: disable=no-init

    @staticmethod
    def get_traceback(exception):
        # -- ASSUMPTION: assert isinstance(exception, Exception)
        return getattr(exception, "__traceback__", None)

    @staticmethod
    def set_traceback(exception, exc_traceback=Unknown):
        assert isinstance(exception, Exception)
        if exc_traceback is Unknown:
            exc_traceback = sys.exc_info()[2]
        exception.__traceback__ = exc_traceback

    @classmethod
    def has_traceback(cls, exception):
        """Indicates if traceback information related to this exception
        is stored with the exception object.

        :param exception:   Exception object to check.
        :return: True, if traceback info is stored. False, otherwise.
        """
        return cls.get_traceback(exception) is not None

    @classmethod
    def describe(cls, exception, use_traceback=False, prefix=""):
        # -- NORMAL CASE:
        text = "{prefix}{0}: {1}\n".format(exception.__class__.__name__,
                                            exception, prefix=prefix)
        if use_traceback:
            exc_traceback = cls.get_traceback(exception)
            if exc_traceback:
                # -- NOTE: Chained-exception cause (see: PEP-3134).
                text += "".join(traceback.format_tb(exc_traceback))
        return text


class ChainedExceptionUtil(ExceptionUtil):
    """Provides a utility class for accessing/modifying exception information
    related to chained exceptions.

    .. seealso:: PEP-3134 Chained excpetions
    """
    # pylint: disable=no-init

    @staticmethod
    def get_cause(exception):
        # -- ASSUMPTION: assert isinstance(exception, Exception)
        return getattr(exception, "__cause__", None)

    @staticmethod
    def set_cause(exception, exc_cause):
        assert isinstance(exception, Exception)
        assert isinstance(exc_cause, Exception) or exc_cause is None
        exception.__cause__ = exc_cause
        if exc_cause and not hasattr(exc_cause, "__traceback__"):
            # -- NEEDED-FOR: Python2
            # Otherwise, traceback formatting tries to access missing attribute.
            exc_cause.__traceback__ = None

    # pylint: disable=arguments-differ
    @classmethod
    def describe(cls, exception, use_traceback=False, prefix="", style="reversed"):
        """Describes an exception, optionally together with its traceback info.
        Also shows information about exception cause (chained exceptions),
        if exists.

        :param exception:       Exception object to describe.
        :param use_traceback:   Indicates if traceback info should be shown.
        :param prefix:          Optional prefix for description text.
        :param style:           Optional style indicator ("reversed", "normal")
        :return: Exception description as text.
        """
        text = ExceptionUtil.describe(exception, use_traceback, prefix)

        # -- STEP: Collect chained exceptions.
        causes = []
        exc_cause = cls.get_cause(exception)
        while exc_cause:
            causes.append(exc_cause)
            exc_cause = cls.get_cause(exc_cause)

        # -- STEP: Describe causes for chained exceptions.
        parts = []
        if style == "normal":
            prefix = "CAUSE: "
            for exc_cause in reversed(causes):
                cause_text = ExceptionUtil.describe(exc_cause, use_traceback,
                                                    prefix)
                parts.append(cause_text)
                if len(parts) == 1:
                    prefix = "CAUSES: "
            parts.append(text)
        else:
            parts.append(text)
            for exc_cause in causes:
                cause_text = ExceptionUtil.describe(exc_cause, use_traceback,
                                                    prefix="CAUSED-BY: ")
                parts.append(cause_text)
        return "\n".join(parts)
        # if exc_cause:
        #     cause_text =
        #     text += u"\n" + cause_text
        # return text



"""
This module provides the step matchers functionality that matches a
step definition (as text) with step-functions that implement this step.
"""

import copy
import re

import parse
from behave_bo._types import (
    ChainedExceptionUtil,
    ExceptionUtil,
)
from behave_bo.model_core import (
    Argument,
    FileLocation,
    Replayable,
)


# -----------------------------------------------------------------------------
# SECTION: Exceptions
# -----------------------------------------------------------------------------
class StepParseError(ValueError):
    """Exception class, used when step matching fails before a step is run.
    This is normally the case when an error occurs during the type conversion
    of step parameters.
    """

    def __init__(self, text=None, exc_cause=None):
        if not text and exc_cause:
            text = str(exc_cause)

        ValueError.__init__(self, text)
        if exc_cause:
            # -- CHAINED EXCEPTION (see: PEP 3134)
            ChainedExceptionUtil.set_cause(self, exc_cause)


# -----------------------------------------------------------------------------
# SECTION: Model Elements
# -----------------------------------------------------------------------------
class Match(Replayable):
    """An parameter-matched *feature file* step name extracted using
    step decorator `parameters`_.

    .. attribute:: func

       The step function that this match will be applied to.

    .. attribute:: arguments

       A list of :class:`~behave_bo.model_core.Argument` instances containing the
       matched parameters from the step name.
    """
    type = "match"

    def __init__(self, func, arguments=None):
        super().__init__()
        self.func = func
        self.arguments = arguments
        self.location = None
        if func:
            self.location = self.make_location(func)

    def __repr__(self):
        if self.func:
            func_name = self.func.__name__
        else:
            func_name = '<no function>'
        return f'<Match {func_name}, {self.location}>'

    def __eq__(self, other):
        if not isinstance(other, Match):
            return False
        return (self.func, self.location) == (other.func, other.location)

    def with_arguments(self, arguments):
        match = copy.copy(self)
        match.arguments = arguments
        return match

    def run(self, context):
        args = []
        kwargs = {}
        for arg in self.arguments:
            if arg.name is not None:
                kwargs[arg.name] = arg.value
            else:
                args.append(arg.value)

        with context.use_with_user_mode():
            self.func(context, *args, **kwargs)

    @staticmethod
    def make_location(step_function):
        """Extracts the location information from the step function and
        builds a FileLocation object with (filename, line_number) info.

        :param step_function: Function whose location should be determined.
        :return: FileLocation object for step function.
        """
        return FileLocation.for_function(step_function)


class NoMatch(Match):
    """Used for an "undefined step" when it can not be matched with a
    step definition.
    """

    def __init__(self):
        Match.__init__(self, func=None)
        self.func = None
        self.arguments = []
        self.location = None


class MatchWithError(Match):
    """Match class when error occur during step-matching

    REASON:
      * Type conversion error occured.
      * ...
    """

    def __init__(self, func, error):
        if not ExceptionUtil.has_traceback(error):
            ExceptionUtil.set_traceback(error)
        Match.__init__(self, func=func)
        self.stored_error = error

    def run(self, context):
        """Raises stored error from step matching phase (type conversion)."""
        raise StepParseError(exc_cause=self.stored_error)


# -----------------------------------------------------------------------------
# SECTION: Matchers
# -----------------------------------------------------------------------------
class Matcher:
    """Pull parameters out of step names.

    .. attribute:: pattern

       The match pattern attached to the step function.

    .. attribute:: func

       The step function the pattern is being attached to.
    """
    schema = "@%s('%s')"  # Schema used to describe step definition (matcher)

    def __init__(self, func, pattern, step_type=None):
        self.func = func
        self.pattern = pattern
        self.step_type = step_type
        self._location = None

    @property
    def location(self):
        if self._location is None:
            self._location = Match.make_location(self.func)
        return self._location

    @property
    def regex_pattern(self):
        """Return the used textual regex pattern."""
        # -- ASSUMPTION: pattern attribute provides regex-pattern
        # NOTE: Method must be overridden if assumption is not met.
        return self.pattern

    def describe(self, schema=None):
        """Provide a textual description of the step function/matcher object.

        :param schema:  Text schema to use.
        :return: Textual description of this step definition (matcher).
        """
        step_type = self.step_type or "step"
        if not schema:
            schema = self.schema
        return schema % (step_type, self.pattern)

    def check_match(self, step):
        """Match me against the "step" name supplied.

        Return None, if I don't match otherwise return a list of matches as
        :class:`~behave_bo.model_core.Argument` instances.

        The return value from this function will be converted into a
        :class:`~behave_bo.matchers.Match` instance by *behave_bo*.
        """
        raise NotImplementedError

    def match(self, step):
        # -- PROTECT AGAINST: Type conversion errors (with ParseMatcher).
        try:
            result = self.check_match(step)
        except Exception as e:  # pylint: disable=broad-except
            return MatchWithError(self.func, e)

        if result is None:
            return None  # -- NO-MATCH
        return Match(self.func, result)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.pattern!r}>"


class ParseMatcher(Matcher):
    """Uses :class:`~parse.Parser` class to be able to use simpler
    parse expressions compared to normal regular expressions.
    """
    custom_types = {}
    parser_class = parse.Parser

    def __init__(self, func, pattern, step_type=None):
        super().__init__(func, pattern, step_type)
        self.parser = self.parser_class(pattern, self.custom_types)

    @property
    def regex_pattern(self):
        # -- OVERWRITTEN: Pattern as regex text.
        return self.parser._expression  # pylint: disable=protected-access

    def check_match(self, step):
        # -- FAILURE-POINT: Type conversion of parameters may fail here.
        #    NOTE: Type converter should raise ValueError in case of PARSE ERRORS.
        result = self.parser.parse(step)
        if not result:
            return None

        args = []
        for index, value in enumerate(result.fixed):
            start, end = result.spans[index]
            args.append(Argument(start, end, step[start:end], value))
        for name, value in result.named.items():
            start, end = result.spans[name]
            args.append(Argument(start, end, step[start:end], value, name))
        args.sort(key=lambda x: x.start)
        return args


class CaseSensitiveParser(parse.Parser):
    """Парсер шагов с учётом регистра."""

    def __init__(self, *args, **kwargs):
        super().__init__(case_sensitive=True, *args, **kwargs)


def register_type(**kw):
    # pylint: disable=anomalous-backslash-in-string
    # REQUIRED-BY: code example
    r"""Registers a custom type that will be available to "parse"
    for type conversion during step matching.

    Converters should be supplied as ``name=callable`` arguments (or as dict).

    A type converter should follow :pypi:`parse` module rules.
    In general, a type converter is a function that converts text (as string)
    into a value-type (type converted value).

    EXAMPLE:

    .. code-block:: python

        from behave_bo import register_type, given
        import parse

        # -- TYPE CONVERTER: For a simple, positive integer number.
        @parse.with_pattern(r"\d+")
        def parse_number(text):
            return int(text)

        # -- REGISTER TYPE-CONVERTER: With behave_bo
        register_type(Number=parse_number)

        # -- STEP DEFINITIONS: Use type converter.
        @given('{amount:Number} vehicles')
        def step_impl(context, amount):
            assert isinstance(amount, int)
    """
    ParseMatcher.custom_types.update(kw)


class RegexMatcher(Matcher):
    def __init__(self, func, pattern, step_type=None):
        super().__init__(func, pattern, step_type)
        self.regex = re.compile(self.pattern)

    def check_match(self, step):
        m = self.regex.match(step)
        if not m:
            return None

        groupindex = {y: x for x, y in self.regex.groupindex.items()}
        args = []
        for index, group in enumerate(m.groups()):
            index += 1
            name = groupindex.get(index, None)
            args.append(Argument(m.start(index), m.end(index), group,
                                 group, name))

        return args


class SimplifiedRegexMatcher(RegexMatcher):
    """Simplified regular expression step-matcher that automatically adds
    start-of-line/end-of-line matcher symbols to string:

    .. code-block:: python

        @when(u'a step passes')     # re.pattern = "^a step passes$"
        def step_impl(context): pass
    """

    def __init__(self, func, pattern, step_type=None):
        assert not (pattern.startswith("^") or pattern.endswith("$")), \
            "Regular expression should not use begin/end-markers: " + pattern
        expression = "^%s$" % pattern
        super().__init__(func, expression, step_type)
        self.pattern = pattern


class CucumberRegexMatcher(RegexMatcher):
    """Compatible to (old) Cucumber style regular expressions.
    Text must contain start-of-line/end-of-line matcher symbols to string:

    .. code-block:: python

        @when(u'^a step passes$')   # re.pattern = "^a step passes$"
        def step_impl(context): pass
    """


matcher_mapping = {
    "parse": ParseMatcher,
    "re": SimplifiedRegexMatcher,

    # -- BACKWARD-COMPATIBLE REGEX MATCHER: Old Cucumber compatible style.
    # To make it the default step-matcher use the following snippet:
    #   # -- FILE: features/environment.py
    #   from behave_bo import use_step_matcher
    #   def before_all(context):
    #       use_step_matcher("re0")
    "re0": CucumberRegexMatcher,
}
current_matcher = ParseMatcher  # pylint: disable=invalid-name


def use_step_matcher(name):
    """Change the parameter matcher used in parsing step text.

    The change is immediate and may be performed between step definitions in
    your step implementation modules - allowing adjacent steps to use different
    matchers if necessary.

    There are several parsers available in *behave_bo* (by default):

    **parse** (the default, based on: :pypi:`parse`)
        Provides a simple parser that replaces regular expressions for
        step parameters with a readable syntax like ``{param:Type}``.
        The syntax is inspired by the Python builtin ``string.format()``
        function.
        Step parameters must use the named fields syntax of :pypi:`parse`
        in step definitions. The named fields are extracted,
        optionally type converted and then used as step function arguments.

        Supports type conversions by using type converters
        (see :func:`~behave_bo.register_type()`).

    **re**
        This uses full regular expressions to parse the clause text. You will
        need to use named groups "(?P<name>...)" to define the variables pulled
        from the text and passed to your ``step()`` function.

        Type conversion is **not supported**.
        A step function writer may implement type conversion
        inside the step function (implementation).

    You may `define your own matcher`_.

    .. _`define your own matcher`: api.html#step-parameters
    """
    global current_matcher  # pylint: disable=global-statement
    current_matcher = matcher_mapping[name]


def get_matcher(func, pattern):
    return current_matcher(func, pattern)



"""
Provides a step registry and step decorators.
The step registry allows to match steps (model elements) with
step implementations (step definitions). This is necessary to execute steps.
"""

from behave_bo.matchers import (
    Match,
    get_matcher,
)
from behave_bo.textutil import (
    text as _text,
)


# limit import * to just the decorators
# pylint: disable=undefined-all-variable
# names = "given when then step"
# names = names + " " + names.title()
# __all__ = names.split()
__all__ = [
    "given", "when", "then", "step", "step_template",  # PREFERRED.
    "Given", "When", "Then", "Step", "Step_Template"  # Also possible.
]

GIVEN = "given"
WHEN = "when"
THEN = "then"
STEP_TEMPLATE = "step_template"
STEP = "step"

step_types = (GIVEN, WHEN, THEN, STEP_TEMPLATE, STEP,)
uniq_step_types = (STEP,)


class AmbiguousStep(ValueError):
    pass


class StepRegistry:
    def __init__(self):
        self.steps = {
            GIVEN: [],
            WHEN: [],
            THEN: [],
            STEP_TEMPLATE: [],
            STEP: {},  # uniq steps
        }

    @staticmethod
    def same_step_definition(step, other_pattern, other_location):
        return (step.pattern == other_pattern and
                step.location == other_location and
                other_location.filename != "<string>")

    def base_add_step_definition(self, keyword, step_text, func):
        step_location = Match.make_location(func)
        step_type = keyword.lower()
        step_text = _text(step_text)
        step_definitions = self.steps[step_type]

        for existing in step_definitions:
            if self.same_step_definition(existing, step_text, step_location):
                # -- EXACT-STEP: Same step function is already registered.
                # This may occur when a step module imports another one.
                return
            elif existing.match(step_text):  # -- SIMPLISTIC
                message = "%s has already been defined in\n  existing step %s"
                new_step = f"@{step_type}('{step_text}')"
                existing.step_type = step_type
                existing_step = existing.describe()
                existing_step += " at %s" % existing.location
                raise AmbiguousStep(message % (new_step, existing_step))

        step_definitions.append(get_matcher(func, step_text))

    def base_find_step_definition(self, step):
        candidates = self.steps[step.step_type]
        more_steps = self.steps[STEP_TEMPLATE]

        if step.step_type != STEP_TEMPLATE and more_steps:
            # -- ENSURE: self.step_type lists are not modified/extended.
            candidates = list(candidates)
            candidates += more_steps

        for step_definition in candidates:
            if step_definition.match(step.name):
                return step_definition

        return None

    def base_find_match(self, step):
        candidates = self.steps[step.step_type]
        more_steps = self.steps[STEP_TEMPLATE]

        if step.step_type != STEP_TEMPLATE and more_steps:
            # -- ENSURE: self.step_type lists are not modified/extended.
            candidates = list(candidates)
            candidates += more_steps

        for step_definition in candidates:
            result = step_definition.match(step.name)
            if result:
                return result

        return None

    def add_step_definition(self, keyword, step_text, func):
        step_type = keyword.lower()
        step_location = Match.make_location(func)
        step_text = _text(step_text)
        step_definitions = self.steps[step_type]

        if step_type in uniq_step_types:
            # Обрабатываем как уникальное определение шага
            if step_text not in step_definitions:
                step_definitions[step_text] = get_matcher(func, step_text)
            else:
                if not self.same_step_definition(step_definitions[step_text], step_text, step_location):
                    raise AmbiguousStep(f'existing step\n\t{step_text}')
        else:
            self.base_add_step_definition(keyword, step_text, func)
        
    def find_step_definition(self, step):
        try:
            step_definition = self.steps[STEP][step.name]
        except KeyError:
            step_definition = self.base_find_step_definition(step)

        return step_definition

    def find_match(self, step):
        try:
            step_match = self.steps[STEP][step.name].match(step.name)
        except KeyError:
            step_match = self.base_find_match(step)

        return step_match

    def make_decorator(self, step_type):
        def decorator(step_text):
            def wrapper(func):
                self.add_step_definition(step_type, step_text, func)
                return func
            return wrapper
        return decorator


registry = StepRegistry()


# -- Create the decorators
# pylint: disable=redefined-outer-name
def setup_step_decorators(run_context=None, registry=registry):
    if run_context is None:
        run_context = globals()
    for step_type in step_types:
        step_decorator = registry.make_decorator(step_type)
        run_context[step_type.title()] = run_context[step_type] = step_decorator


# -----------------------------------------------------------------------------
# MODULE INIT:
# -----------------------------------------------------------------------------
setup_step_decorators()



import io
from typing import (
    Any,
    Dict,
)

from django.http.response import (
    HttpResponse,
)

from behave_bo.enums import (
    StepTypeEnum,
)
from behave_bo.behave_django_bo.signals import (
    skip_request_check_signal,
)


class AppRequest:
    """
    Запрос к приложению
    """

    def __init__(
        self,
        path: str,
        method: str,
        context,
        parameters: Dict[str, Any],
        force: bool = False,
    ):
        self._path = path
        self._method = method
        self._context = context
        self._parameters = parameters
        self._force = force

    def execute(self) -> HttpResponse:
        """
        Выполнение запроса к приложению
        """
        method = getattr(self._context.session, self._method.lower())

        skip_request = self.check_skip_request()

        if self._force or not skip_request:
            response = method(
                path=self._path,
                data=self._parameters
            )
            is_used = True
        else:
            response = HttpResponse()
            is_used = False

        if self._context.config.remove_readonly_requests:
            self._context.config.django_test_runner.collect_requests_data(is_used)

        for parameter in self._parameters.values():
            if isinstance(parameter, io.IOBase):
                parameter.close()
            elif isinstance(parameter, list):
                for item in parameter:
                    if isinstance(item, io.IOBase):
                        item.close()

        return response

    def check_skip_request(self) -> bool:
        """
        Проверка параметра пропуска запроса
        """
        skip_request = False

        if (
            (self._context.config.skip_readonly_requests or self._context.config.remove_readonly_requests)
            and hasattr(self._context, 'current_step') and self._context.current_step.step_type != StepTypeEnum.THEN
        ):
            if self._context.config.skip_readonly_requests_url_endings:
                skip_url_endings = self._context.config.skip_readonly_requests_url_endings.split(',')
            else:
                # TODO Указать адреса для списка ниже на уровне web-bb-core
                skip_url_endings = [
                    '/rows',
                    '/edit',
                    '/list',
                    '/select',
                    '/check_notify',
                ]

            if any(self._path.endswith(url) for url in skip_url_endings) and self.check_skip_request_signal():
                skip_request = True

        return skip_request

    def check_skip_request_signal(self):
        results = skip_request_check_signal.send(sender=self)
        return all(res for _, res in results)



from django.conf import (
    settings,
)


def get_context_extractor_class():
    # TODO BOBUH-23698 Вынести настройку CONTEXT_EXTRACTOR из web_bb_behave
    context_extractor_cls = settings.WEB_BB_BEHAVE__CONTEXT_EXTRACTOR
    context_extractor_cls_path = context_extractor_cls.split('.')
    # Allow for relative paths
    if len(context_extractor_cls_path) > 1:
        extractor_module_name = '.'.join(context_extractor_cls_path[:-1])
    else:
        extractor_module_name = '.'
    extractor_module = __import__(extractor_module_name, {}, {}, context_extractor_cls_path[-1])
    return getattr(extractor_module, context_extractor_cls_path[-1])


context_extractor_class = get_context_extractor_class()
context_extractor = context_extractor_class()



import multiprocessing
import sys

from behave_bo.behave_django_bo.enums import (
    SaveDbParameterEnum,
)
from behave_bo.behave_django_bo.utils import (
    get_features_paths,
)
from behave_bo.configuration import (
    options as behave_options,
)
from django.conf import (
    settings,
)
from django.core.cache import (
    DEFAULT_CACHE_ALIAS,
    cache,
)
from django.core.management.base import (
    BaseCommand,
)
from django.test.signals import (
    clear_cache_handlers,
)
from django.test.utils import (
    get_runner,
)


def set_default_cache_backend():
    """Указание BehaveLocMemCache в качестве cache-backend по-молчанию если задана настройка."""
    if settings.WEB_BB_BEHAVE__USE_LOC_MEM_CACHE:
        settings.CACHES[DEFAULT_CACHE_ALIAS] = {
            'BACKEND': 'behave_bo.behave_django_bo.cache.BehaveLocMemCache',
            'LOCATION': '',
            'OPTIONS': {
                'MAX_ENTRIES': 500,
            },
        }


def add_command_arguments(parser):
    """
    Additional command line arguments for the behave management command

    @param parser: ArgumentParser для парсинга аргументов команды
    """
    parser.add_argument(
        '--use-existing-database',
        action='store_true',
        default=False,
        help="Don't create a test database. USE AT YOUR OWN RISK!",
    )
    parser.add_argument(
        '-k', '--keepdb',
        action='store_true',
        default=False,
        help="Preserves the test DB between runs.",
    )
    parser.add_argument(
        '-S', '--simple',
        action='store_true',
        default=False,
        help="Use simple test runner that supports Django's"
             " testing client only (no web browser automation)"
    )
    parser.add_argument(
        '--all',
        action='store_true',
        default=False,
        help='Discover tests in all apps and plugins and run them',
    )
    parser.add_argument(
        '--plugins',
        action='store',
        metavar='PLUGINS',
        dest='plugins',
        help='Enable runner plugins (comma separated)',
    )
    parser.add_argument(
        '--plugin_vars',
        action='store',
        metavar='VARS',
        dest='plugin_vars',
        help='Plugin variables',
    )
    parser.add_argument(
        '--disable_failed_plugin',
        action='store_true',
        dest='disable_failed_plugin',
        help='Disables plugin if it failed',
    )
    parser.add_argument(
        '--parallel',
        action='store_true',
        default=False,
        help='Run behave tests using up to N parallel processes, which can be set with --parallel-count.',
    )
    parser.add_argument(
        '--parallel-count',
        default=multiprocessing.cpu_count(),
        type=int,
        help='Parallel processes count.',
    )
    parser.add_argument(
        '--save-db',
        action='store',
        choices=SaveDbParameterEnum.values,
        help='Save changes by autotest to default db or separate db',
    )
    parser.add_argument(
        '--bgjob-wait-time',
        action='store',
        default=180,
        type=int,
        help='BackgroundJob wait cycle timeout in seconds.',
    )
    parser.add_argument(
        '--bgjob-wait-iteration-freeze',
        action='store',
        default=0.5,
        type=float,
        help='BackgroundJob wait iteration delay in seconds.',
    )
    parser.add_argument(
        '--generate-fixture',
        action='store',
        default='off',
        help='Generate fixtures for tests: "off": don\'t generate fixtures; '
             '"generate-only": only generate fixtures '
             'without execute behave tests; '
             '"generate": generate fixtures and execute behave tests.',
    )
    parser.add_argument(
        '--remove-readonly-requests',
        action='store_true',
        dest='remove_readonly_requests',
        default=False,
        help='Remove readonly requests files.',
    )
    parser.add_argument(
        '--skip-readonly-requests',
        action='store_true',
        default=False,
        help='Skip sending of readonly request (AppRequest) for non-check steps.',
    )
    parser.add_argument(
        '--skip-readonly-requests-url-endings',
        action='store',
        default='',
        help='Url endings for skip request (comma separated). Eg. /rows,/list',
    )
    parser.add_argument(
        '--skip-production-request-checks',
        action='store_true',
        default=False,
        help='Skip checking of production-request data.',
    )
    parser.add_argument(
        '--fix-checks',
        action='store_true',
        default=False,
        help='Fix failed autotests checks.',
    )
    parser.add_argument(
        '--fix-excluded-keys',
        action='store',
        default=None,
        help='Add all different keys to excluded_keys param.',
    )
    parser.add_argument(
        '--optimize-steps-loading',
        action='store_true',
        default=False,
        help='Using cached step locations. Subsequent runs are much faster '
             'for scenarios with already cached steps. Cleanup is recommended '
             'when changes are made to the location of step implementations. '
             'Use option "--clear-cached-step-locations" to clean up.',
    )
    parser.add_argument(
        '--optimize-features-steps-loading',
        action='store',
        default='',
        help='Указать путь до файла в котором будет сформирован кэш step_definitions по feature-файлам.'
             'Файл формируется при запуске без --parallel, а используется при запуске с --parallel.',
    )
    parser.add_argument(
        '--clear-cached-step-locations',
        action='store_true',
        default=False,
        help='Clearing cached step locations.',
    )
    parser.add_argument(
        '--analyze-model-records',
        action='store',
        default='',
        help='Use "ModelRecordAnalyzer" tool for specified scenarios.',
    )
    parser.add_argument(
        '--in-tags-order',
        action='store_true',
        default=False,
        help='Run scenarios in the order of specified tags in --tags.',
    )
    parser.add_argument(
        '--rerun-if-failed',
        action='store',
        default=0,
        type=int,
        help='Rerun failed tests if the number of failed tests does not exceed '
             'the specified value. 0 means rerun will not occur.',
    )
    parser.add_argument(
        '--measure-requests-coverage',
        action='store_true',
        dest='measure_requests_coverage',
        default=False,
        help='Enable context switching per request.',
    )
    parser.add_argument(
        '--log-non-200-requests',
        action='store_true',
        default=False,
        dest='log_non_200_requests',
        help='Enable logging requests with non 200 status code.',
    )
    parser.add_argument(
        '--collect-top-scenarios',
        action='store_true',
        default=False,
        dest='collect_top_scenarios',
        help='Create top_scenarios.html in junit_directory',
    )


def add_behave_arguments(parser):  # noqa
    """
    Additional command line arguments extracted directly from behave
    """

    # Option strings that conflict with Django
    conflicts = [
        '--no-color',
        '--version',
        '-c',
        '-k',
        '-v',
        '-S',
        '--simple',
    ]

    parser.add_argument(
        'paths',
        action='store',
        nargs='*',
        help="Feature directory, file or file location (FILE:LINE)."
    )

    for fixed, keywords in behave_options:
        keywords = keywords.copy()

        # Configfile only entries are ignored
        if not fixed:
            continue

        # Build option strings
        option_strings = []
        for option in fixed:
            # Prefix conflicting option strings with `--behave`
            if option in conflicts:
                prefix = '--' if option.startswith('--') else '-'
                option = option.replace(prefix, '--behave-', 1)

            option_strings.append(option)

        # config_help isn't a valid keyword for add_argument
        if 'config_help' in keywords:
            keywords['help'] = keywords['config_help']
            del keywords['config_help']

        parser.add_argument(*option_strings, **keywords)


class Command(BaseCommand):
    help = 'Runs behave tests'
    requires_system_checks = False

    def add_arguments(self, parser):
        """
        Add behave's and our command line arguments to the command
        """
        parser.usage = "%(prog)s [options] [ [DIR|FILE|FILE:LINE] ]+"
        parser.description = """\
        Run a number of feature tests with behave_bo."""

        add_command_arguments(parser)
        add_behave_arguments(parser)

    def handle(self, *args, **options):
        set_default_cache_backend()

        # импортируем класс раннера, который указан в настройках
        django_test_runner_cls = get_runner(settings)

        django_test_runner = django_test_runner_cls(**options)
        django_test_runner.setup_test_environment()

        django_test_runner.collect_reset_sequences_sql()

        behave_args = self.get_behave_args()

        feature_paths = []

        if options['all']:
            feature_paths = get_features_paths()

        if not feature_paths:
            sys.exit()

        behave_args.extend(feature_paths)

        self.prerun_checks(options)

        django_test_runner.setup_plugins(behave_args, options)

        exit_status = django_test_runner.run_behave_tests(behave_args, options)

        # Teardown django environment
        django_test_runner.teardown_test_environment()

        if exit_status != 0:
            sys.exit(exit_status)

    def prerun_checks(self, options):
        """
        Выполнить проверки перед запуском
        """
        try:
            cache_backend = settings.CACHES[DEFAULT_CACHE_ALIAS]['BACKEND']
            cache_backend = cache_backend.rpartition('.')[2]
        except KeyError:
            cache_backend = None

        if options['parallel'] and cache_backend == 'RedisCache':
            redis_db_count = cache.client.connect().config_get('databases')['databases']
            clear_cache_handlers(setting='CACHES')
            cpu_thread_count = multiprocessing.cpu_count()

            if int(redis_db_count) < cpu_thread_count:
                raise Exception(
                    'При параллельном запуске тестов с использование redis в качестве кэша '
                    'количество бд redis должно быть больше или равно количеству потоков процессора.\n'
                    f'redis_db_count: {redis_db_count}, cpu_thread_count: {cpu_thread_count}'
                )

    def get_behave_args(self, argv=sys.argv):
        """
        Get a list of those command line arguments specified with the
        management command that are meant as arguments for running behave_bo.
        """
        parser = BehaveArgsHelper().create_parser('manage.py', 'behave')
        args, unknown = parser.parse_known_args(argv[2:])

        behave_args = []
        for option in unknown:
            # Remove behave prefix
            if option.startswith('--behave-'):
                option = option.replace('--behave-', '', 1)
                prefix = '-' if len(option) == 1 else '--'
                option = prefix + option

            behave_args.append(option)

        return behave_args


class BehaveArgsHelper(Command):

    def add_arguments(self, parser):
        """
        Override setup of command line arguments to make behave commands not
        be recognized. The unrecognized args will then be for behave! :)
        """
        add_command_arguments(parser)



"""behave_bo is behaviour-driven development, Python style

Behavior-driven development (or BDD) is an agile software development
technique that encourages collaboration between developers, QA and
non-technical or business participants in a software project.

*behave_bo* uses tests written in a natural language style, backed up by Python
code.

To get started, we recommend the `tutorial`_ and then the `test language`_ and
`api`_ references.

.. _`tutorial`: tutorial.html
.. _`test language`: gherkin.html
.. _`api`: api.html
"""

from behave_bo.fixture import (
    fixture,
    use_fixture,
)
from behave_bo.matchers import (
    register_type,
    use_step_matcher,
)
from behave_bo.step_registry import *  # pylint: disable=wildcard-import


# pylint: disable=undefined-all-variable
__all__ = [
    "given", "when", "then", "step", "step_template", "use_step_matcher", "register_type",
    "Given", "When", "Then", "Step", "Step_Template",
    "fixture", "use_fixture",
]
__version__ = "1.2.6"



# STATUS: Basic concept works.
"""
A **fixture** provides a concept to simplify test support functionality
that needs a setup/cleanup cycle per scenario, feature or test-run.
A fixture is provided as fixture-function that contains the setup part and
cleanup part similar to :func:`contextlib.contextmanager` or `pytest.fixture`_.

.. _pytest.fixture: https://docs.pytest.org/en/latest/fixture.html

A fixture is used when:

* the (registered) fixture tag is used for a scenario or feature
* the :func:`.use_fixture()` is called in the environment file (normally)

.. sourcecode:: python

    # -- FILE: behave4my_project/fixtures.py (or: features/environment.py)
    from behave_bo import fixture
    from somewhere.browser.firefox import FirefoxBrowser

    @fixture
    def browser_firefox(context, timeout=30, **kwargs):
        # -- SETUP-FIXTURE PART:
        context.browser = FirefoxBrowser(timeout, *args, **kwargs)
        yield context.browser
        # -- CLEANUP-FIXTURE PART:
        context.browser.shutdown()

.. sourcecode:: gherkin

    # -- FILE: features/use_fixture.feature
    Feature: Use Fixture in Scenario

        @fixture.browser.firefox
        Scenario: Use browser=firefox
          Given I use the browser
          ...
        # -- AFTER-SCENEARIO: Cleanup fixture.browser.firefox

.. sourcecode:: python

    # -- FILE: features/environment.py
    from behave_bo import use_fixture
    from behave4my_project.fixtures import browser_firefox

    def before_tag(context, tag):
        if tag == "fixture.browser.firefox":
            # -- Performs fixture setup and registers fixture cleanup
            use_fixture(browser_firefox, context, timeout=10)

.. hidden:

    BEHAVIORAL DECISIONS:

    * Should scenario/feature be executed when fixture-setup fails
      (similar to before-hook failures) ?
      NO, scope is skipped, but after-hooks and cleanups are executed.

    * Should remaining fixture-setups be performed after first fixture fails?
      NO, first setup-error aborts the setup and execution of the scope.

    * Should remaining fixture-cleanups be performed when first cleanup-error
      occurs?
      YES, try to perform all fixture-cleanups and then reraise the
      first cleanup-error.


    OPEN ISSUES:

    * AUTO_CALL_REGISTERED_FIXTURE (planned in future):
        Run fixture setup before or after before-hooks?

    IDEAS:

    * Fixture registers itself in fixture registry (runtime context).
    * Code in before_tag() will either be replaced w/ fixture processing function
      or will be automatically be executed (AUTO_CALL_REGISTERED_FIXTURE)
    * Support fixture tags w/ parameters that are automatically parsed and
      passed to fixture function, like:
      @fixture(name="foo", pattern="{name}={browser}")
"""

import inspect


# -------------------------------------------------------------------------------
# LOCAL HELPERS:
# -------------------------------------------------------------------------------
def iscoroutinefunction(func):
    """Checks if a function is a coroutine-function, like:

     * ``async def f(): ...`` (since Python 3.5)
     * ``@asyncio.coroutine def f(): ...`` (since Python3)

    .. note:: Compatibility helper

        Avoids to import :mod:`asyncio` module directly (since Python3),
        which in turns initializes the :mod:`logging` module as side-effect.

    :param func:  Function to check.
    :return: True, if function is a coroutine function.
             False, otherwise.
    """
    # -- NOTE: inspect.iscoroutinefunction() is available since Python 3.5
    #    Checks also if @asyncio.coroutine decorator is not used.
    # pylint: disable=no-member
    return (getattr(func, "_is_coroutine", False) or
            (hasattr(inspect, "iscoroutinefunction") and
             inspect.iscoroutinefunction(func)))


def is_context_manager(func):
    """Checks if a fixture function provides context-manager functionality,
    similar to :func`contextlib.contextmanager()` function decorator.

    .. code-block:: python

        @fixture
        def foo(context, *args, **kwargs):
            context.foo = setup_foo()
            yield context.foo
            cleanup_foo()

        @fixture
        def bar(context, *args, **kwargs):
            context.bar = setup_bar()
            return context.bar

        assert is_context_manager(foo) is True      # Generator-function
        assert is_context_manager(bar) is False     # Normal function

    :param func:    Function to check.
    :return: True, if function is a generator/context-manager function.
             False, otherwise.
    """
    genfunc = inspect.isgeneratorfunction(func)
    return genfunc and not iscoroutinefunction(func)


# -------------------------------------------------------------------------------
# EXCEPTIONS:
# -------------------------------------------------------------------------------
class InvalidFixtureError(RuntimeError):
    """Raised when a fixture is invalid.
    This occurs when a generator-function with more than one yield statement
    is used as fixture-function.
    """


# -------------------------------------------------------------------------------
# FUNCTIONS: Fixture support
# -------------------------------------------------------------------------------
def _setup_fixture(fixture_func, context, *fixture_args, **fixture_kwargs):
    """Provides core functionality to setup a fixture and registers its
    cleanup part (if needed).
    """
    if is_context_manager(fixture_func):
        # -- CASE: Fixture function is a two-step generator (setup, cleanup).
        def cleanup_fixture():
            try:
                next(func_it)           # CLEANUP-FIXTURE PART
                # -- USE func_it: From outer scope (here).
            except StopIteration:
                return False
            # -- NOT-NEEDED:
            # except Exception:
            #    raise   # -- CLEANUP-FIXTURE PART raised an error.
            else:
                message = "Has more than one yield: %r" % fixture_func
                raise InvalidFixtureError(message)

        # -- GENERATOR: Get instance via call and register cleanup.
        #  Then perform setup_fixture to ensure that cleanup_fixture()
        #  is called even if setup-error occurs.
        #  NOTE: cleanup_fixture() is called when context layer is removed.
        func_it = fixture_func(context, *fixture_args, **fixture_kwargs)
        context.add_cleanup(cleanup_fixture)
        setup_result = next(func_it) # SETUP-FIXTURE PART (may raise error)
    else:
        # -- CASE: Fixture is a simple function (setup-only)
        # NOTE: No cleanup is registered (not needed by intention of user)
        setup_result = fixture_func(context, *fixture_args, **fixture_kwargs)
    return setup_result


def use_fixture(fixture_func, context, *fixture_args, **fixture_kwargs):
    """Use fixture (function) and call it to perform its setup-part.

    The fixture-function is similar to a :func:`contextlib.contextmanager`
    (and contains a yield-statement to seperate setup and cleanup part).
    If it contains a yield-statement, it registers a context-cleanup function
    to the context object to perform the fixture-cleanup at the end of the
    current scoped when the context layer is removed
    (and all context-cleanup functions are called).

    Therefore, fixture-cleanup is performed after scenario, feature or test-run
    (depending when its fixture-setup is performed).

    .. code-block:: python

        # -- FILE: behave4my_project/fixtures.py (or: features/environment.py)
        from behave_bo import fixture
        from somewhere.browser import FirefoxBrowser

        @fixture(name="fixture.browser.firefox")
        def browser_firefox(context, *args, **kwargs):
            # -- SETUP-FIXTURE PART:
            context.browser = FirefoxBrowser(*args, **kwargs)
            yield context.browser
            # -- CLEANUP-FIXTURE PART:
            context.browser.shutdown()

    .. code-block:: python

        # -- FILE: features/environment.py
        from behave_bo import use_fixture
        from behave4my_project.fixtures import browser_firefox

        def before_tag(context, tag):
            if tag == "fixture.browser.firefox":
                use_fixture(browser_firefox, context, timeout=10)


    :param fixture_func: Fixture function to use.
    :param context: Context object to use
    :param fixture_kwargs: Positional args, passed to the fixture function.
    :param fixture_kwargs: Additional kwargs, passed to the fixture function.
    :return: Setup result object (may be None).
    """
    return _setup_fixture(fixture_func, context, *fixture_args, **fixture_kwargs)


def use_fixture_by_tag(tag, context, fixture_registry):
    """Process any fixture-tag to perform :func:`use_fixture()` for its fixture.
    If the fixture-tag is known, the fixture data is retrieved from the
    fixture registry.

    .. code-block:: python

        # -- FILE: features/environment.py
        from behave_bo.fixture import use_fixture_by_tag
        from behave4my_project.fixtures import browser_firefox, browser_chrome

        # -- SCHEMA 1: fixture_func
        fixture_registry1 = {
            "fixture.browser.firefox": browser_firefox,
            "fixture.browser.chrome":  browser_chrome,
        }
        # -- SCHEMA 2: fixture_func, fixture_args, fixture_kwargs
        fixture_registry2 = {
            "fixture.browser.firefox": (browser_firefox, (), dict(timeout=10)),
            "fixture.browser.chrome":  (browser_chrome,  (), dict(timeout=12)),
        }

        def before_tag(context, tag):
            if tag.startswith("fixture."):
                return use_fixture_by_tag(tag, context, fixture_registry1):
            # -- MORE: Tag processing steps ...


    :param tag:     Fixture tag to process.
    :param context: Runtime context object, used for :func:`use_fixture()`.
    :param fixture_registry:  Registry maps fixture-tag to fixture data.
    :return: Fixture-setup result (same as: use_fixture())
    :raises LookupError: If fixture-tag/fixture is unknown.
    :raises ValueError: If fixture data type is not supported.
    """
    fixture_data = fixture_registry.get(tag, None)
    if fixture_data is None:
        raise LookupError("Unknown fixture-tag: %s" % tag)

    if callable(fixture_data):
        fixture_func = fixture_data
        use_fixture(fixture_func, context)
    elif isinstance(fixture_data, (tuple, list)):
        assert len(fixture_data) == 3
        fixture_func, fixture_args, fixture_kwargs = fixture_data
        return use_fixture(fixture_func, context, *fixture_args, **fixture_kwargs)
    else:
        message = "fixture_data: Expected tuple or fixture-func, but is: %r"
        raise ValueError(message % fixture_data)


def fixture_call_params(fixture_func, *args, **kwargs):
    # -- SEE: use_composite_fixture_with()
    return (fixture_func, args, kwargs)


def use_composite_fixture_with(context, fixture_funcs_with_params):
    """Helper function when complex fixtures should be created and
    safe-cleanup is needed even if an setup-fixture-error occurs.

    This function ensures that fixture-cleanup is performed
    for every fixture that was setup before the setup-error occured.

    .. code-block:: python

        # -- BAD-EXAMPLE: Simplistic composite-fixture
        # NOTE: Created fixtures (fixture1) are not cleaned up.
        @fixture
        def foo_and_bad0(context, *args, **kwargs):
            the_fixture1 = setup_fixture_foo(*args, **kwargs)
            the_fixture2 = setup_fixture_bar_with_error("OOPS-HERE")
            yield (the_fixture1, the_fixture2)  # NOT_REACHED.
            # -- NOT_REACHED: Due to fixture2-setup-error.
            the_fixture1.cleanup()  # NOT-CALLED (SAD).
            the_fixture2.cleanup()  # OOPS, the_fixture2 is None (if called).

    .. code-block:: python

        # -- GOOD-EXAMPLE: Sane composite-fixture
        # NOTE: Fixture foo.cleanup() occurs even after fixture2-setup-error.
        @fixture
        def foo(context, *args, **kwargs):
            the_fixture = setup_fixture_foo(*args, **kwargs)
            yield the_fixture
            cleanup_fixture_foo(the_fixture)

        @fixture
        def bad_with_setup_error(context, *args, **kwargs):
            raise RuntimeError("BAD-FIXTURE-SETUP")

        # -- SOLUTION 1: With use_fixture()
        @fixture
        def foo_and_bad1(context, *args, **kwargs):
            the_fixture1 = use_fixture(foo, context, *args, **kwargs)
            the_fixture2 = use_fixture(bad_with_setup_error, context, "OOPS")
            return (the_fixture1, the_fixture2) # NOT_REACHED

        # -- SOLUTION 2: With use_composite_fixture_with()
        @fixture
        def foo_and_bad2(context, *args, **kwargs):
            the_fixture = use_composite_fixture_with(context, [
                fixture_call_params(foo, *args, **kwargs),
                fixture_call_params(bad_with_setup_error, "OOPS")
             ])
            return the_fixture

    :param context:     Runtime context object, used for all fixtures.
    :param fixture_funcs_with_params: List of fixture functions with params.
    :return: List of created fixture objects.
    """
    composite_fixture = []
    for fixture_func, args, kwargs in fixture_funcs_with_params:
        the_fixture = use_fixture(fixture_func, context, *args, **kwargs)
        composite_fixture.append(the_fixture)
    return composite_fixture



# -------------------------------------------------------------------------------
# DECORATORS:
# -------------------------------------------------------------------------------
def fixture(func=None, name=None, pattern=None):
    """Fixture decorator (currently mostly syntactic sugar).

    .. code-block:: python

        # -- FILE: features/environment.py
        # CASE FIXTURE-GENERATOR-FUNCTION (like @contextlib.contextmanager):
        @fixture
        def foo(context, *args, **kwargs):
            the_fixture = setup_fixture_foo(*args, **kwargs)
            context.foo = the_fixture
            yield the_fixture
            cleanup_fixture_foo(the_fixture)

        # CASE FIXTURE-FUNCTION: No cleanup or cleanup via context-cleanup.
        @fixture(name="fixture.bar")
        def bar(context, *args, **kwargs):
            the_fixture = setup_fixture_bar(*args, **kwargs)
            context.bar = the_fixture
            context.add_cleanup(cleanup_fixture_bar, the_fixture.cleanup)
            return the_fixture

    :param name:    Specifies the fixture tag name (as string).

    .. seealso::

        * :func:`contextlib.contextmanager` decorator
        * `@pytest.fixture`_
    """
    def mark_as_fixture(func, name=None, pattern=None):
        func.name = name
        func.pattern = pattern
        func.behave_fixture = True
        return func

    if func is None:
        # CASE: @fixture()
        # CASE: @fixture(name="foo")
        def decorator(func):
            return mark_as_fixture(func, name, pattern=pattern)
        return decorator
    elif callable(func):
        # CASE: @fixture
        return mark_as_fixture(func, name, pattern=pattern)
    else:
        # -- OOPS: Should be callable-object or None
        message = f"Invalid func: func={func!r}, name={name!r}"
        raise TypeError(message)



from django.conf import (
    settings,
)
from django.db.models.query_utils import (
    DeferredAttribute,
)
from django.test.runner import (
    DiscoverRunner,
)
from django.test.utils import (
    override_settings,
)

from behave_bo.__main__ import (
    run_behave,
)
from behave_bo.behave_django_bo.consts import (
    override_settings_prefix,
)
from behave_bo.behave_django_bo.utils import (
    override_dirs,
)
from behave_bo.configuration import (
    Configuration,
)
from behave_bo.loggers import (
    tests_logger,
)
from behave_bo.model import (
    Scenario,
)
from behave_bo.reporter.failed_tags import (
    FailedTagsParallelReporter,
    FailedTagsReporter,
)
from behave_bo.reporter.junit import (
    JUnitReporter,
)
from behave_bo.runner import (
    BehaveRunner,
    ParallelBehaveRunner,
    BehaveTagsCachedRunner,
)


behave_run_scenario = Scenario.run


class PatchedScenario(Scenario):

    @property
    def override_settings_tag(self):
        override_settings_tag = None

        for tag in self.tags:
            if tag.startswith(override_settings_prefix):
                override_settings_tag = tag
                break

        return override_settings_tag

    def make_override_settings_dict(self, runner) -> dict:
        """Формирует словарь настроек для переопределения в override_settings.

        Настройки, задаваемые в runner.override_settings_dict, будут переопределены
        или дополнены для конкретного сценария.

        Args:
            runner: объект класса Runner, исполнителя behave-автотестов.

        Returns:
            Словарь вида {Наименование настройки: значение}.

        """
        override_settings_dict = getattr(runner, 'override_settings_dict', {})

        if self.override_settings_tag:
            settings_string = self.override_settings_tag.removeprefix(override_settings_prefix)
            for key, value in map(lambda s: s.split('=', 1), settings_string.split(',')):
                if value.isdigit():
                    value = int(value)
                elif value.lower() in ('true', 'false'):
                    value = value.lower() == 'true'
                elif value.lower() == 'none':
                    value = None

                override_settings_dict[key] = value

            tests_logger.info(
                f'Для сценария {self.main_tag} будут переопределены django settings: {override_settings_dict}'
            )

        return override_settings_dict

    def run_scenario(self, runner):
        """Пропатченный метод запуска behave-сценария.

        Сразу проверяется необходимость выполнения сценария, и если не нужно выполнять, помечается пропущенным.
        Если нужно выполнять, на время выполнения конкретного сценария создаёт уникальную временную директорию.

        Args:
            self: объект класса Scenario
            runner: объект класса Runner, исполнителя behave-автотестов.
        """
        failed = False
        should_run_scenario = self.should_run(runner.config)

        if should_run_scenario:
            for number, step in enumerate(self.steps, 1):
                step.number = number
            settings_dict = self.make_override_settings_dict(runner)
            with override_dirs(self), override_settings(**settings_dict):
                failed = behave_run_scenario(self, runner)
        else:
            self.mark_skipped()

        return failed


Scenario.base_skip_tags = list(Scenario.base_skip_tags) + [override_settings_prefix]
Scenario.override_settings_tag = PatchedScenario.override_settings_tag
Scenario.make_override_settings_dict = PatchedScenario.make_override_settings_dict
Scenario.run = PatchedScenario.run_scenario


class BaseBehaveTestRunner(DiscoverRunner):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.return_code_success = 0
        self.behave_config = None

        if kwargs.get('save_db'):
            # В случае если нужно сохранить результат в БД / отдельную БД,
            # атрибут keepdb должен быть True
            self.keepdb = True

    def get_tempdir_path(self):
        raise NotImplementedError

    def get_template_data_types(self):
        raise NotImplementedError

    def get_environment_filepath(self):
        raise NotImplementedError

    def run_behave_tests(self, behave_args, options):
        """Выполняет запуск behave тестов.

        Args:
            behave_args: Аргументы, относящиеся к behave_bo.
            options: Все аргументы команды запуска.

        Returns:
            Результат выполнения тестов.

        """
        self.behave_config = Configuration(behave_args)
        try:
            log_capture_exclude_template = settings.WEB_BB_BEHAVE__LOG_CAPTURE_EXCLUDE_TEMPLATE
        except AttributeError:
            log_capture_exclude_template = ''

        self.behave_config.log_capture_exclude_template = log_capture_exclude_template

        self._add_extra_parameters_to_behave_config(options)

        if self.parallel:
            behave_runner_class = ParallelBehaveRunner
            failed_tags_reporter_class = FailedTagsParallelReporter
        else:
            if self.behave_config.optimize_steps_loading:
                behave_runner_class = BehaveTagsCachedRunner
            else:
                behave_runner_class = BehaveRunner
            failed_tags_reporter_class = FailedTagsReporter

        failed_tags_reporter = failed_tags_reporter_class(self.behave_config)
        self.behave_config.reporters.append(failed_tags_reporter)

        try:
            junit_reporter = [r for r in self.behave_config.reporters if isinstance(r, JUnitReporter)][0]
        except IndexError:
            junit_reporter = None

        runner_kwargs = {
            'tempdir': self.get_tempdir_path(),
            'step_template_data_types': self.get_template_data_types(),
            'environment_filepath': self.get_environment_filepath(),
        }

        run_result = run_behave(
            self.behave_config,
            runner_class=behave_runner_class,
            runner_kwargs=runner_kwargs,
        )

        if failed_tags_reporter.rerun_required():
            tests_logger.info('Перезапуск упавших сценариев для выявления ложных падений...\n')
            failed_tags_reporter.rerun_up()

            if junit_reporter:
                junit_reporter.rerun = True
                self.behave_config.reporters.append(junit_reporter)

            run_behave(
                self.behave_config,
                runner_class=behave_runner_class,
                runner_kwargs=runner_kwargs,
            )

            if failed_tags_reporter.check_all_false_positive():
                run_result = self.return_code_success

        return run_result

    def _add_extra_parameters_to_behave_config(self, options):
        """Добавляет дополнительные параметры в конфиг в соответствии с options.

        Args:
            options: Аргументы команды запуска.

        """
        config = self.behave_config
        config.django_test_runner = self

        config.parallel = options.get('parallel')
        config.parallel_count = options.get('parallel_count')
        config.collect_top_scenarios = options.get('collect_top_scenarios')
        config.save_db = options.get('save_db')
        config.optimize_steps_loading = options.get('optimize_steps_loading')
        config.clear_cached_step_locations = options.get('clear_cached_step_locations')
        config.optimize_features_steps_loading = options.get('optimize_features_steps_loading')
        config.in_tags_order = options.get('in_tags_order')
        config.rerun_if_failed = options.get('rerun_if_failed')
        config.measure_requests_coverage = options.get('measure_requests_coverage')

        config.save_changes_to_db = bool(config.save_db)

        return config


def _DeferredAttributeWithCheck__get__(self, instance, cls=None):
    """
    Добавление проверки/логирования в метод отложенного получения значения атрибута.

    Args:
        self: экземпляр класса DeferredAttribute
        instance: экземпляр класса модели для которого выполняется получение значения атрибута.
        cls: класс модели

    Returns:
        Значение атрибута.

    Raises:
        Исключение в случае если включена настройка CHECK_DEFERRED_ATTR_GET
         и значение атрибута не найдено в кэше, а значит его планируется получить из БД выполнив дополнительный запрос.
    """
    if instance is None:
        return self
    data = instance.__dict__
    field_name = self.field.attname

    if field_name in data:
        return data[field_name]

    attr = f"{instance.__class__.__name__}.{field_name}"
    message = f"Lazy fetching of {attr} may cause 1+N issue"

    if settings.WEB_BB_BEHAVE__CHECK_DEFERRED_ATTR_GET:
        raise AssertionError(message)
    elif settings.WEB_BB_BEHAVE__LOG_DEFERRED_ATTR_GET:
        tests_logger.warning(message)

    return DeferredAttribute__get(self, instance, cls)


DeferredAttribute__get, DeferredAttribute.__get__ = DeferredAttribute.__get__, _DeferredAttributeWithCheck__get__



import re
import time
from datetime import (
    datetime,
)

from django.core.cache import (
    DEFAULT_CACHE_ALIAS,
    cache,
    caches,
)
from django.core.cache.backends.locmem import (
    LocMemCache,
)
from django_redis.cache import (
    RedisCache,
)

from behave_bo.loggers import (
    tests_logger,
)


# Сохранять TTL при дампе DjangoCache с типом бэекенда Redis
DUMP_REDIS_CACHE_WITH_TTL = False


class BehaveLocMemCache(LocMemCache):
    """Cache для использования в тестах. Дополнен методом keys,
    аналогичным методу keys в RedisCache"""

    def keys(self, pattern):
        """
        Метод для получения списка ключей по шаблону из кеша приложения.
        Возвращает ключи без служебного префикса, который добавляется по-умолчанию в методах LocMemCache.

        Args:
            pattern: Регулярное выражение для поиска ключей в кэше
        """
        pattern = pattern.replace('*', '.*')
        res = list()

        with self._lock:
            cache_keys = self._cache.copy()

        for key in cache_keys:
            # получим изначальный ключ удалив из него префикс key_prefix:version:
            pure_key = key.replace(f'{self.key_prefix}:{self.version}:', '', 1)

            if re.match(pattern, pure_key) and not self._has_expired(key):
                res.append(pure_key)

        return res

    def ttl(self, key, version=None):
        """Получение значение ttl для ключа.

        Эмулирует метод django_redis.client.default.DefaultClient.ttl.

        Args:
            key: Значение ключа в кэше;
            version: Версия ключа в кэше.

        Returns:
            ttl ключа или None, если ключ не существует.
        """
        time_to_live = 0

        if self.has_key(key, version=version):
            key = self.make_key(key, version=version)

            if expire_time := self._expire_info.get(key, None):
                time_to_live = int(expire_time - time.time())
            else:
                time_to_live = None

        return time_to_live

    def _cull(self):
        """Удаляет ключи при превышении их количества, задаваемого в настройке MAX_ENTRIES.

        Здесь просто выведем в лог сообщение для отслеживания вызова метода.
        """
        tests_logger.warn(
            f'Превышен лимит количества хранимых ключей, равный {self._max_entries}. '
            f'{len(self._cache) // self._cull_frequency} ключей будет удалено.'
        )
        super()._cull()


def get_cache_backend():
    """"
        Возвращает класс backend'а Django Cache
    """
    cache_backend = caches[DEFAULT_CACHE_ALIAS].__class__

    return cache_backend


def setup_cache_server_db(db_num):
    """
        Устанавливает используемую в redis бд путем замены окончания в переменной _server.
    """
    cache_backend = get_cache_backend()
    if issubclass(cache_backend, RedisCache):
        server_url = cache._server.rsplit('/', 1)[0]
        cache._server = f'{server_url}/{db_num}'


def django_cache_dump():
    """
        Функция сохраняет текущие значения ключей из Django Cache
        и возвращает их в словаре. Формат возвращаемого словаря зависит от
        backend'а которы используется для Django Cache
    """
    cache_backend = get_cache_backend()
    cache_dump = {}
    if issubclass(cache_backend, LocMemCache):
        with cache._lock:
            cache_dump['_cache'] = cache._cache.copy()
            cache_dump['_expire_info'] = cache._expire_info.copy()
    elif issubclass(cache_backend, RedisCache):
        keys = cache.keys('*')
        cache_dump['redis_keys'] = cache.get_many(keys)
        if DUMP_REDIS_CACHE_WITH_TTL:
            cache_dump['redis_ttl'] = {}
            for key in keys:
                cache_dump['redis_ttl'][key] = cache.ttl(key)
            cache_dump['time'] = datetime.now()

    return cache_dump


def django_cache_restore(cache_dump):
    """
        Функция принимает на вход снимок Django Cache,
        полученный в django_cache_dump(), и восстанавливает состояние кеша
        на основании этого снимка.
    """
    cache_backend = get_cache_backend()
    if issubclass(cache_backend, LocMemCache):
        with cache._lock:
            if all(k in cache_dump for k in ('_cache', '_expire_info')):
                cache._cache = cache_dump.get('_cache')
                cache._expire_info = cache_dump.get('_expire_info')
    elif issubclass(cache_backend, RedisCache):
        cache.clear()
        if not DUMP_REDIS_CACHE_WITH_TTL:
            cache.set_many(cache_dump.get('redis_keys', {}))
        else:
            time_passed = (datetime.now() - cache_dump['time']).total_seconds()
            for key, value in cache_dump['redis_keys'].items():
                old_ttl = cache_dump['redis_ttl'].get(key, 0)
                new_ttl = int(max(old_ttl - time_passed, 0))
                cache.add(key, value, timeout=new_ttl)
